const request = require('request-promise-native')
const fs = require('fs')
const _url = require('url')
const _path = require('path')

/**
 * Class that allows you to communicate with the Umbraco ModelsBuilder Api.
 */
class ApiClient {

  /**
   * Create a new Api Client.
   * @param {string} url - Umbraco site URL.
   * @param {string} user - Umbraco user.
   * @param {string} password - Umbraco password.
   * @param {string} path - Models path.
   */
  constructor (url, user, password, path) {
    this._url = _url.parse(url).href
    this._user = user
    this._pass = password
    this._path = _path.resolve(path)
    this._request = request.defaults({
      baseUrl: this._url,
      auth: {
        user: this._user,
        password: this._pass,
        sendImmediately: true,
      },
    })
  }

  /**
   * Send a request that validates the client version.
   * @returns Promise<*>
   * @private
   */
  _validateClientVersion () {
    return this._request({
      uri: '/Umbraco/BackOffice/ModelsBuilder/ModelsBuilderApi/ValidateClientVersion',
      method: 'POST',
      body: {
        ClientVersionString: '3.0.0.0',
        MinServerVersionSupportingClientString: '3.0.0.0',
      },
      json: true,
    })
  }

  /**
   * Get all files in the path that end with '.cs'.
   * @returns {Promise<string[]>}
   * @private
   */
  _getAllFiles () {
    return new Promise((resolve, reject) => {
      fs.readdir(this._path, (err, files) => {
        if (err) {
          return reject(err)
        }

        resolve(files.filter(f => /\.cs$/.test(f))
          .map(f => _path.join(this._path, f)))
      })
    })
  }

  /**
   * Generates a file delete task.
   * @param {string} file - File to delete.
   * @returns {Promise<string>}
   * @private
   */
  _deleteTask (file) {
    return new Promise((resolve, reject) => {
      fs.unlink(file, err => {
        if (err) {
          return reject(err)
        }
        resolve(file)
      })
    })
  }

  /**
   * Delete all files ending with .generated.cs.
   * @param {string[]} files - List of files.
   * @returns {Promise<string[]>}
   * @private
   */
  _deleteFiles (files) {
    const tasks = files.filter(f => /\.generated\.cs$/.test(f))
      .map(f => this._deleteTask(f))

    return Promise.all(tasks)
  }

  /**
   * Fetch the generated models from the server.
   * @param {string} namespace - The model namespace.
   * @param {Map<string, string>} files - The file map.
   * @returns Promise<*>
   * @private
   */
  _fetchModels (namespace, files) {
    return this._request({
      uri: '/Umbraco/BackOffice/ModelsBuilder/ModelsBuilderApi/GetModels',
      method: 'POST',
      body: {
        Namespace: namespace,
        Files: files,
        ClientVersionString: '3.0.0.0',
        MinServerVersionSupportingClientString: '3.0.0.0',
      },
      json: true,
    })
  }

  /**
   * Generates a file read task.
   * @param {string} file - The file to read.
   * @returns {Promise<[string]>}
   * @private
   */
  _readTask (file) {
    return new Promise((resolve, reject) => {
      fs.readFile(file, {encoding: 'utf-8'}, (err, contents) => {
        if (err) {
          return reject(err)
        }
        resolve([file, contents || ''])
      })
    })
  }

  /**
   * Read all files that do **not** end with '.generated.cs'.
   * @param {string[]} files - File list.
   * @returns {Promise<string[][]>}
   * @private
   */
  _readFiles (files) {
    const tasks = files.filter(f => /^(.(?!.generated))*\.cs$/.test(f))
      .map(f => this._readTask(f))

    return Promise.all(tasks)
  }

  /**
   * Read files and transform them into a map.
   * @param {string[]} files - File list.
   * @returns {Promise<Map<string, string>>}
   * @private
   */
  async _mapFiles (files) {
    return (await this._readFiles(files))
      .reduce((map, entry) => {
        map.set(_path.parse(entry[0]).name, entry[1])
        return map
      }, new Map())
  }

  /**
   * Generate a file save task for a specific model.
   * @param {string} modelName - The model's name.
   * @param {string} modelContents - The model's code.
   * @returns {Promise<string>}
   * @private
   */
  _saveTask (modelName, modelContents) {
    return new Promise((resolve, reject) => {
      const modelPath = _path.join(this._path, `${modelName}.generated.cs`)
      fs.writeFile(modelPath, modelContents, err => {
        if (err) {
          return reject(err)
        }
        resolve(modelPath)
      })
    })
  }

  /**
   * Save all models to disk.
   * @param {Map<string, string>} models - The models map.
   * @returns {Promise<string[]>}
   * @private
   */
  _saveModels (models) {
    const tasks = Object.keys(models)
      .map(k => this._saveTask(k, models[k]))

    return Promise.all(tasks)
  }

  /**
   * Delete generated models, fetch newly generated models from the server and save them to the disk.
   * @param {string} namespace - The model namespace.
   * @returns {Promise<string[]>}
   */
  async getModels (namespace) {
    const files = await this._getAllFiles()
    await this._deleteFiles(files)

    await this._validateClientVersion()

    const fileMap = await this._mapFiles(files)
    const models = await this._fetchModels(namespace, fileMap)
    return await this._saveModels(models)
  }
}

module.exports = ApiClient